==> vortex_array-0.31.0.tar.gz: vortex_array-0.31.0/vortex-array/src/compute/fill_null.rs <==
use vortex_error::{VortexExpect, VortexResult, vortex_bail};
use vortex_scalar::Scalar;

use crate::compute::try_cast;
use crate::encoding::Encoding;
use crate::{Array, ArrayRef, IntoArray};

/// Implementation of fill_null for an encoding.
///
/// SAFETY: the fill value is guaranteed to be non-null.
pub trait FillNullFn<A> {
    fn fill_null(&self, array: A, fill_value: Scalar) -> VortexResult<ArrayRef>;
}

impl<E: Encoding> FillNullFn<&dyn Array> for E
where
    E: for<'a> FillNullFn<&'a E::Array>,
{
    fn fill_null(&self, array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
        let array_ref = array
            .as_any()
            .downcast_ref::<E::Array>()
            .vortex_expect("Failed to downcast array");
        FillNullFn::fill_null(self, array_ref, fill_value)
    }
}

pub fn fill_null(array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
    if !array.dtype().is_nullable() {
        return Ok(array.to_array());
    }

    if array.invalid_count()? == 0 {
        return try_cast(array, fill_value.dtype());
    }

    if fill_value.is_null() {
        vortex_bail!("Cannot fill_null with a null value")
    }

    if !array.dtype().eq_ignore_nullability(fill_value.dtype()) {
        vortex_bail!(MismatchedTypes: array.dtype(), fill_value.dtype())
    }

    let fill_value_nullability = fill_value.dtype().nullability();
    let filled = fill_null_impl(array, fill_value)?;

    assert_eq!(
        filled.len(),
        array.len(),
        "FillNull length mismatch {}",
        array.encoding()
    );
    assert_eq!(
        filled.dtype().nullability(),
        fill_value_nullability,
        "FillNull nullability mismatch {}",
        array.encoding()
    );

    Ok(filled)
}

fn fill_null_impl(array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
    if let Some(fill_null_fn) = array.vtable().fill_null_fn() {
        return fill_null_fn.fill_null(array, fill_value);
    }

    log::debug!("FillNullFn not implemented for {}", array.encoding());
    let canonical_arr = array.to_canonical()?.into_array();
    if let Some(fill_null_fn) = canonical_arr.vtable().fill_null_fn() {
        return fill_null_fn.fill_null(&canonical_arr, fill_value);
    }

    vortex_bail!(
        "fill null not implemented for canonical encoding {}, fallback from {}",
        canonical_arr.encoding(),
        array.encoding()
    )
}

==> vortex_array-0.32.0.tar.gz: vortex_array-0.32.0/vortex-array/src/compute/fill_null.rs <==
use vortex_error::{VortexExpect, VortexResult, vortex_bail};
use vortex_scalar::Scalar;

use crate::compute::cast;
use crate::encoding::Encoding;
use crate::{Array, ArrayRef, IntoArray};

/// Implementation of fill_null for an encoding.
///
/// SAFETY: the fill value is guaranteed to be non-null.
pub trait FillNullFn<A> {
    // TODO(ngates): take &Scalar
    fn fill_null(&self, array: A, fill_value: Scalar) -> VortexResult<ArrayRef>;
}

impl<E: Encoding> FillNullFn<&dyn Array> for E
where
    E: for<'a> FillNullFn<&'a E::Array>,
{
    fn fill_null(&self, array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
        let array_ref = array
            .as_any()
            .downcast_ref::<E::Array>()
            .vortex_expect("Failed to downcast array");
        FillNullFn::fill_null(self, array_ref, fill_value)
    }
}

pub fn fill_null(array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
    if !array.dtype().is_nullable() {
        return Ok(array.to_array());
    }

    if array.invalid_count()? == 0 {
        return cast(array, fill_value.dtype());
    }

    if fill_value.is_null() {
        vortex_bail!("Cannot fill_null with a null value")
    }

    if !array.dtype().eq_ignore_nullability(fill_value.dtype()) {
        vortex_bail!(MismatchedTypes: array.dtype(), fill_value.dtype())
    }

    let fill_value_nullability = fill_value.dtype().nullability();
    let filled = fill_null_impl(array, fill_value)?;

    assert_eq!(
        filled.len(),
        array.len(),
        "FillNull length mismatch {}",
        array.encoding()
    );
    assert_eq!(
        filled.dtype().nullability(),
        fill_value_nullability,
        "FillNull nullability mismatch {}",
        array.encoding()
    );

    Ok(filled)
}

fn fill_null_impl(array: &dyn Array, fill_value: Scalar) -> VortexResult<ArrayRef> {
    if let Some(fill_null_fn) = array.vtable().fill_null_fn() {
        return fill_null_fn.fill_null(array, fill_value);
    }

    log::debug!("FillNullFn not implemented for {}", array.encoding());
    let canonical_arr = array.to_canonical()?.into_array();
    if let Some(fill_null_fn) = canonical_arr.vtable().fill_null_fn() {
        return fill_null_fn.fill_null(&canonical_arr, fill_value);
    }

    vortex_bail!(
        "fill null not implemented for canonical encoding {}, fallback from {}",
        canonical_arr.encoding(),
        array.encoding()
    )
}
